from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework import get_version
from dodo_commands.framework.config import Paths
from dodo_commands.framework.util import bordered
from semantic_version import Version
from plumbum import local
import os
import ruamel.yaml
import sys


def _args():
    parser = ArgumentParser(description='Compare configuration version to '
                            'version in original project config file.')
    parser.add_argument(
        '--dodo',
        dest='check_dodo',
        action='store_true',
        help=("Check that the installed dodo commands version satisfies the "
              "minimal dodo commands version in the dodo config "
              "(/ROOT/required_dodo_commands_version)"))
    parser.add_argument(
        '--config',
        dest='check_config',
        action='store_true',
        help=("Check that the version field in the local dodo commands config "
              "(/ROOT/version) is up-to-date with the version in the shared "
              "dodo commands config."))
    args = Dodo.parse_args(parser)
    return args


def _shared_config_dir():
    src_dir = Dodo.get_config("/ROOT/src_dir", None)
    default_shared_config_dir = os.path.join(src_dir, "extra", "dodo_commands",
                                             "res") if src_dir else None
    return Dodo.get_config("/ROOT/shared_config_dir",
                           default_shared_config_dir)


def _get_root_config_field(config_filename, field_name):
    with open(config_filename) as f:
        config = ruamel.yaml.round_trip_load(f.read())
    return config.get("ROOT", {}).get(field_name, "")


def _partial_sem_version(version):
    v = Version(version)
    return Version("%s.%s" % (v.major, v.minor), partial=True)


def _config_filename():
    return os.path.join(Paths().res_dir(), "config.yaml")


def check_dodo_commands_version():  # noqa
    required_version = _get_root_config_field(
        _config_filename(), 'required_dodo_commands_version')
    if required_version:
        actual_version = get_version()
        if Version(actual_version) < Version(required_version):
            sys.stdout.write(
                bordered(
                    'The dodo_commands package needs to be upgraded (%s < %s). Tip: use "dodo upgrade"'
                    % (
                        actual_version,
                        required_version,
                    ), ))
            sys.stdout.write('\n')


def check_config_version():  # noqa
    if not _shared_config_dir():
        return

    original_file = os.path.join(_shared_config_dir(), "config.yaml")
    if not os.path.exists(original_file):
        return

    original_version = _get_root_config_field(original_file, 'version')
    if not original_version:
        sys.stderr.write(
            "No version found in original file %s\n" % original_file)
        return

    copied_file = os.path.join(Paths().res_dir(), "config.yaml")
    copied_version = _get_root_config_field(copied_file, 'version')
    if not copied_version:
        sys.stderr.write(
            "No version found in user managed file %s\n" % copied_file)
        return

    if (_partial_sem_version(copied_version) <
            _partial_sem_version(original_version)):
        sys.stdout.write(
            bordered(
                'Configuration needs update (%s < %s). Tip: use "dodo diff ."\n'
                % (
                    copied_version,
                    original_version,
                )))
        sys.stdout.write('\n')


def check_config_changes():  # noqa
    if os.path.exists(os.path.join(Paths().res_dir(), '.git')):
        with local.cwd(Paths().res_dir()):
            changes = local['git']('status', '--short')
            if changes:
                sys.stdout.write(
                    bordered(
                        'Configuration has changes (tip: use "dodo commit-config"):\n'
                        + changes))
                sys.stdout.write('\n')


if Dodo.is_main(__name__, safe=False):
    args = _args()
    if args.check_dodo:
        check_dodo_commands_version()
    if args.check_config:
        check_config_version()
        check_config_changes()



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.util import filter_choices
from plumbum.cmd import docker
from six.moves import input as raw_input


def _args():
    parser = ArgumentParser()
    args = Dodo.parse_args(parser)
    return args


def _containers():
    result = []
    for line in docker("ps", "--format",
                       "{{.ID}} {{.Names}} {{.Image}}").split('\n'):
        if line:
            cid, name, image = line.split()
            result.append(dict(name=name, cid=cid, image=image))
    return result


if Dodo.is_main(__name__):
    args = _args()
    containers = _containers()

    for idx, container in enumerate(containers):
        print("%d - %s" % (idx + 1, container['name']))

    raw_choice = raw_input("Select a container: ")

    selected_containers, span = filter_choices(containers, raw_choice)
    if span == [0, len(raw_choice)]:
        for container in selected_containers:
            Dodo.run(['docker', 'kill', container['cid']], )
    else:
        raise CommandError("Syntax error")



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
import os
import sys
from six.moves import input as raw_input
from dodo_commands.framework.util import filter_choices
from plumbum.cmd import tmux

session_id = os.path.expandvars('$USER')


def _args():
    parser = ArgumentParser()
    parser.add_argument('--kill-session', action='store_true')
    parser.add_argument('--list', action='store_true')
    parser.add_argument('--run', type=int, nargs='?', const=-1)
    args = Dodo.parse_args(parser)
    args.command_map = Dodo.get_config('/TMUX/commands', [])
    return args


def _create_tmux_window():
    # Create tmux session
    Dodo.run(['tmux', '-2', 'new-session', '-d', '-s', session_id], )

    # Create a tmux window
    Dodo.run(['tmux', 'new-window', '-t', '%s:1' % session_id, '-n', 'Logs'], )


def _get_commands_and_labels(command_map):
    label_size = 0
    for label in command_map:
        label_size = max(label_size, len(label))
    label_prefix = "%0" + str(label_size) + "s"

    commands, labels = [], []
    for label in command_map:
        for command in command_map[label]:
            commands.append(command)
            format_string = "%02s [" + label_prefix + "] - %s"
            labels.append(format_string % (str(len(commands)), label, command))

    return commands, labels


def _get_selected_commands(commands, labels):
    print()
    for label in labels:
        print(label)

    raw_choice = raw_input(
        '\nSelect one or more commands (e.g. 1,3-4), or type a command: ')
    selected_commands, span = filter_choices(commands, raw_choice)
    if span != [0, len(raw_choice)]:
        selected_commands = [raw_choice]
    return selected_commands


if Dodo.is_main(__name__):
    args = _args()
    check_exists = Dodo.get_config('/TMUX/check_exists', '/')
    if not os.path.exists(check_exists):
        raise CommandError("Path %s does not exist" % check_exists)

    commands, labels = _get_commands_and_labels(args.command_map)

    if args.list:
        print()
        for label in labels:
            print(label)
        sys.exit(0)

    if args.run:
        selected_commands = (_get_selected_commands(commands, labels)
                             if args.run == -1 else [commands[args.run - 1]])
        for command in selected_commands:
            Dodo.run(['bash', '-c', command])
        sys.exit(0)

    has_session = False
    try:
        sessions = tmux('ls')
        for session in sessions.split('\n'):
            has_session = has_session or session.startswith('%s:' % session_id)
    except:
        pass

    if has_session and args.kill_session:
        Dodo.run(['tmux', 'kill-session', '-t', session_id], )

    if not has_session:
        _create_tmux_window()
        Dodo.run(['tmux', 'send-keys', 'dodo tmux', 'C-m'], )
        # Attach to tmux session
        Dodo.run(['tmux', '-2', 'attach-session', '-t', session_id], )
    else:
        while True:
            selected_commands = _get_selected_commands(commands, labels)
            for command in selected_commands:
                Dodo.run(['tmux', 'split-window', '-v'], )
                Dodo.run(['tmux', 'send-keys', command, 'C-m'], )

            # Set default window
            tmux('select-pane', '-t', '0')
            tmux('select-layout', 'tile')



"""Pauses the execution."""
import time


class Decorator:  # noqa
    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            '--pause-ms',
            type=int,
            help="Pause in milliseconds before continuing")

    def modify_args(self, dodo_args, root_node, cwd):  # noqa
        if getattr(dodo_args, 'pause_ms', 0):
            time.sleep(dodo_args.pause_ms / 1000)
        return root_node, cwd



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework.config import Paths
import os


def _args():
    parser = ArgumentParser()
    parser.add_argument(
        '--message', '-m', dest='message', help='The commit message')
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if not os.path.exists(os.path.join(Paths().res_dir(), '.git')):
        Dodo.run(['git', 'init'], cwd=Paths().res_dir())

    Dodo.run(['git', 'add', '-A'], cwd=Paths().res_dir())
    Dodo.run(['git', 'commit', '-m', args.message or 'Update configuration'],
             cwd=Paths().res_dir())



from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.util import EnvironMemo
import os
import re
import sys


class KeyNotFound(CommandError):
    """Raised if an xpath is not found in a configuration."""
    pass


def _xpath_string(node):
    return '/' + '/'.join([str(x) for x in node.xpath])


class DictKey:
    def __init__(self, the_dict, the_key, xpath):
        self.dict = the_dict
        self.key = the_key
        self.xpath = xpath

    def __repr__(self):  # noqa
        return "DK[%s]" % self.key

    def warning_msg_not_expanded(self):
        return "Unexpanded key {key} at location /{xpath}".format(
            key=self.key, xpath='/'.join([str(x) for x in self.xpath]))

    def create_dict_val(self, expanded_key):  # noqa
        if expanded_key != self.key:
            self.dict[expanded_key] = self.dict[self.key]
            del self.dict[self.key]
        return DictVal(self.dict, expanded_key, self.xpath)


class DictVal:
    def __init__(self, the_dict, the_key, xpath):
        self.dict = the_dict
        self.key = the_key
        self.xpath = xpath

    def get_value(self):
        return self.dict[self.key]

    def replace_value(self, new_value):
        self.dict[self.key] = new_value

    def warning_msg_not_expanded(self):
        return "Unexpanded value {val} at location /{xpath}".format(
            val=self.get_value(), xpath='/'.join([str(x) for x in self.xpath]))

    def __repr__(self):  # noqa
        return "DV[%s]" % self.key


class ListVal:
    def __init__(self, the_list, the_idx, xpath):
        self.idx = the_idx
        self.list = the_list
        self.xpath = xpath

    def get_value(self):
        return self.list[self.idx]

    def replace_value(self, new_value):
        self.list[self.idx] = new_value

    def warning_msg_not_expanded(self):
        return "Unexpanded value {val} at location /{xpath}".format(
            val=self.get_value(), xpath='/'.join([str(x) for x in self.xpath]))

    def __repr__(self):  # noqa
        return "L[%s]" % self.idx


class Key:
    """Access a nested value in a dict."""

    def __init__(self, config, xpath):  # noqa
        self.config = config
        # the list of sub-keys along the path to an item in self.config
        self.xpath = xpath

    def child(self, subkey):
        """Return key for child with subkey."""
        return Key(self.config, list(self.xpath) + [subkey])

    def _step_into(self, node, subkey):
        try:
            if isinstance(node, type(dict())):
                return node[subkey]
            if isinstance(node, type(list())):
                return node[int(subkey)]
        except:
            raise KeyNotFound("Cannot get value at %s\n" % self)

    def exists(self):
        config_node = self.config
        for xpath_node in self.xpath:
            if isinstance(config_node, dict) and xpath_node in config_node:
                config_node = config_node[xpath_node]
            elif isinstance(config_node,
                            list) and int(xpath_node) < len(config_node):
                config_node = config_node[int(xpath_node)]
            else:
                return False
        return True

    def get(self):  # noqa
        node = self.config
        for subkey in self.xpath:
            node = self._step_into(node, subkey)
        return node

    def set(self, value):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        node[self.xpath[-1]] = value

    def remove(self):  # noqa
        node = self.config
        for subkey in self.xpath[:-1]:
            node = self._step_into(node, subkey)
        del node[self.xpath[-1]]

    def __repr__(self):  # noqa
        return "/" + "/".join([str(x) for x in self.xpath])


class ConfigExpander:
    """Expand environment variables and references in the config."""

    _key_regexp = r"\$\{(/[^\}]*)\}"

    def __init__(self, extra_vars):
        self.extra_vars = extra_vars

    @classmethod
    def _get_xpath_from_string(cls, xpath_string):
        return tuple([ix for ix in xpath_string.split("/") if ix])

    def _expand_str(self, current_str):
        expanded_str = current_str
        key_expressions = [
            x for x in re.finditer(self._key_regexp, current_str)
        ]
        known_strs = []

        while (key_expressions):
            changed = False
            for key_expression in reversed(key_expressions):
                xpath_string = key_expression.group(1)
                xpath = self._get_xpath_from_string(xpath_string)

                key = Key(self.config, xpath)
                if key.exists():
                    expanded_str = (
                        expanded_str[:key_expression.start()] + str(
                            key.get()) + expanded_str[key_expression.end():])

                    if expanded_str not in known_strs:
                        known_strs.append(expanded_str)
                        changed = True

            if not changed:
                return None

            current_str = expanded_str
            key_expressions = [
                x for x in re.finditer(self._key_regexp, current_str)
            ]

        with EnvironMemo(self.extra_vars):
            expanded_str = os.path.expandvars(expanded_str)

        env_key_regexp = r"\$\{([^\}]*)\}"
        env_key_expressions = [
            x for x in re.finditer(env_key_regexp, expanded_str)
        ]
        return expanded_str if not env_key_expressions else None

    def _expand(self, raw_obj):
        if isinstance(raw_obj, str):
            return self._expand_str(raw_obj)

        if isinstance(raw_obj, list):
            return [self._expand(x) for x in raw_obj]

        if isinstance(raw_obj, dict):
            return {
                self._expand(k): self._expand(v)
                for k, v in raw_obj.items()
            }

        return raw_obj

    def _schedule_children(self, obj, todo, xpath):
        if isinstance(obj, dict):
            for k in obj:
                todo.append(DictKey(obj, k, xpath + [k]))
            return True
        if isinstance(obj, list):
            for idx in range(len(obj)):
                todo.append(ListVal(obj, idx, xpath + [idx]))
            return True

    def run(self, config, callbacks=None):  # noqa
        nodes = []
        self.config = config
        self._schedule_children(config, nodes, [])

        changed = True
        while len(nodes) and changed:
            changed = False
            new_nodes = list()
            for node in nodes:
                if isinstance(node, DictKey):
                    expanded_key = self._expand(node.key)
                    if expanded_key is None:
                        new_nodes.append(node)
                    else:
                        changed = True
                        new_nodes.append(node.create_dict_val(expanded_key))
                elif isinstance(node, DictVal) or isinstance(node, ListVal):
                    value = node.get_value()
                    if self._schedule_children(value, new_nodes, node.xpath):
                        changed = True
                    else:
                        expanded_value = self._expand(value)
                        if expanded_value is None:
                            new_nodes.append(node)
                        else:
                            changed = True

                            node.replace_value(expanded_value)
                            if _xpath_string(node) in (callbacks or {}):
                                callbacks[_xpath_string(node)](expanded_value)
                else:
                    raise CommandError("Should not reach here")

            nodes = new_nodes

        for node in nodes:
            sys.stderr.write("Warning: %s\n" % node.warning_msg_not_expanded())



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.config import Paths
from dodo_commands.framework.util import is_using_system_dodo, symlink
from dodo_commands.framework.config import load_global_config_parser
import sys
import os
import tempfile


def _args():
    parser = ArgumentParser(
        description=("Install command packages into the global " +
                     "commands directory. " + _packages_in_extra_dir()))
    parser.add_argument("paths",
                        nargs='*',
                        help='Create symlinks to these command directories')
    parser.add_argument("--pip",
                        nargs='*',
                        help='Pip install the commands in these packages')
    parser.add_argument("--git",
                        nargs='*',
                        help='Clone a git repo into the commands directory')
    parser.add_argument("--remove",
                        action='store_true',
                        help='Remove commands from the commands directory')
    parser.add_argument("--to-defaults",
                        action='store_true',
                        help='Install into the default commands directory')

    group = parser.add_mutually_exclusive_group()

    group.add_argument(
        "--make-default",
        help=
        'Create a symlink to a global commands package in the default commands directory'
    )
    group.add_argument(
        "--remove-default",
        help=
        'Remove a symlink to a global commands package from the default commands directory'
    )
    args = Dodo.parse_args(parser)
    return args


def _packages_in_extra_dir():
    extra_dir = Paths().extra_dir()
    packages = [
        x for x in os.listdir(extra_dir)
        if os.path.isdir(os.path.join(extra_dir, x)) and not x.startswith('__')
    ]

    if len(packages) == 0:
        return ""

    if len(packages) == 1:
        msg = " The %s package is found automagically " % packages[0]
    else:
        packages[-1] = "and " + packages[-1]
        msg = " The %s packages are found automagically " % ", ".join(packages)

    return (msg + " in the dodo_commands.extra package" +
            ", e.g. the following works: dodo install-default-commands %s." %
            packages[0])


def _report_error(msg):
    sys.stderr.write(msg + os.linesep)


def _remove_package(package, only_from_defaults=False):
    """Install the dir with the default commands."""
    if not only_from_defaults:
        dest_dir = os.path.join(Paths().global_commands_dir(), package)
        if not os.path.exists(dest_dir):
            raise CommandError("Not installed: %s" % dest_dir)
        Dodo.run(['rm', '-rf', dest_dir])

    defaults_dest_dir = os.path.join(Paths().default_commands_dir(), package)
    if os.path.exists(defaults_dest_dir):
        Dodo.run(['rm', defaults_dest_dir])


def _install_package(package, install_commands_function, add_to_defaults):
    """Install the dir with the global commands."""
    dest_dir = os.path.join(Paths().global_commands_dir(), package)
    defaults_dest_dir = os.path.join(Paths().default_commands_dir(), package)

    if add_to_defaults and os.path.exists(defaults_dest_dir):
        _report_error("Error: already installed: %s" % defaults_dest_dir)
        return False

    if not install_commands_function():
        return False

    if add_to_defaults:
        if not os.path.exists(dest_dir):
            _report_error("Error: not found: %s" % dest_dir)
            return False

        if os.name == 'nt' and not args.confirm:
            symlink(dest_dir, defaults_dest_dir)
        else:
            Dodo.run(['ln', '-s', dest_dir, defaults_dest_dir])

    return True


def _install_commands_from_path(path, mv=False):
    """Install the dir with the global commands."""
    dest_dir = os.path.join(Paths().global_commands_dir(),
                            os.path.basename(path))

    if not os.path.exists(path):
        alt_path = os.path.join(Paths().extra_dir(), path)
        if os.path.exists(alt_path):
            path = alt_path
        else:
            _report_error("Error: path not found: %s" % path)
            return False

    if os.path.exists(dest_dir):
        _report_error("Error: already installed: %s" % dest_dir)
        return False

    if mv:
        Dodo.run(['mv', path, dest_dir])
    else:
        try:
            if os.name == 'nt' and not args.confirm:
                symlink(os.path.abspath(path), dest_dir)
            else:
                Dodo.run(['ln', '-s', os.path.abspath(path), dest_dir])
        except:
            _report_error("Error: could not create a symlink in %s." %
                          dest_dir)

    return True


def _install_commands_from_package(package):
    config = load_global_config_parser()
    python_path_parts = os.path.split(
        config.get("settings", "python_interpreter"))
    pip_path_parts = python_path_parts[:-1] + python_path_parts[-1].replace(
        'python', 'pip')
    pip = os.path.join(*pip_path_parts)
    Dodo.run([
        pip, 'install', '--upgrade', '--target',
        Paths().global_commands_dir(), package
    ])

    return True


def _clone_git_repo(repo_path):
    tmp_dir = tempfile.mkdtemp()
    Dodo.run(['git', 'clone', repo_path], cwd=tmp_dir)
    package = os.listdir(tmp_dir)[0]
    return tmp_dir, package


if Dodo.is_main(__name__):
    args = _args()
    if args.pip and not is_using_system_dodo():
        raise CommandError(
            "Please deactivate your dodo project first by running 'deactivate'."
        )

    if args.make_default:
        _install_package(args.make_default, lambda: True, True)
        sys.exit(0)

    if args.remove_default:
        _remove_package(args.remove_default, only_from_defaults=True)
        sys.exit(0)

    if args.paths:
        for path in args.paths:
            package = os.path.basename(path)
            if args.remove:
                _remove_package(package)
            else:
                _install_package(
                    package, lambda: _install_commands_from_path(path),
                    args.to_defaults)

    if args.pip:
        for package in args.pip:
            if args.remove:
                _remove_package(package)
            else:
                _install_package(
                    package, lambda: _install_commands_from_package(package),
                    args.to_defaults)

    if args.git:
        for repo_path in args.git:
            if args.remove:
                raise CommandError(
                    "The --git option is not supported when removing a package."
                    + " Please use dodo install-commands --remove <package>.")

            tmp_dir, package = _clone_git_repo(repo_path)
            _install_package(
                package, lambda: _install_commands_from_path(
                    os.path.join(tmp_dir, package), mv=True), args.to_defaults)
            Dodo.run(['rm', '-rf', tmp_dir])



import argcomplete
from argparse import ArgumentParser
from collections import defaultdict
import os
import sys
from dodo_commands.framework.config import (load_global_config_parser,
                                            get_command_path, ConfigLoader)
from dodo_commands.framework.singleton import Dodo, DecoratorScope, ConfigArg  # noqa
from dodo_commands.framework.command_error import CommandError  # noqa
from dodo_commands.framework.command_map import get_command_map, execute_script


def get_version():  # noqa
    return "0.25.2"


class ManagementUtility(object):
    """Internal helper class for executing commands."""

    def __init__(self):  # noqa
        self.prog_name = os.path.basename(sys.argv[0])

    def main_help_text(self, command_map, commands_only=False):
        """Return the script's main help text, as a string."""
        if command_map is None:
            command_map = get_command_map(get_command_path())

        if commands_only:
            usage = sorted(command_map.keys())
        else:
            usage = [
                "",
                "Version %s (%s --version)." % (get_version(), self.prog_name),
                "Type '%s help <command>' for help on "
                "a specific command." % self.prog_name,
                "Available commands (dodo help --commands):\n",
            ]

            max_name_size = -1
            command_groups = defaultdict(lambda: [])
            for command_name, command_map_item in command_map.items():
                max_name_size = max(max_name_size, len(command_name))
                command_groups[command_map_item.group].append(command_name)

            groups = sorted(list(command_groups.values()),
                            key=lambda x: len(x),
                            reverse=True)
            for group in groups:
                group.append("")

            col_width = max_name_size + 10
            rows, cols = os.popen('stty size', 'r').read().split()
            nr_cols = max(1, int(int(cols) / col_width))

            head, tail = groups[:nr_cols], groups[nr_cols:]
            while any([x is not None for x in head]):
                names = []
                for group_idx, group in enumerate(head):
                    names.append('' if group is None else group.pop(0))
                    if group is not None and len(group) == 0:
                        head[group_idx] = (tail.pop(0) if tail else None)
                usage.append("".join(name.ljust(col_width) for name in names))

        return '\n'.join(usage)

    def _handle_exception(self, e):
        if '--traceback' in sys.argv or not isinstance(e, CommandError):
            raise
        sys.stderr.write('%s: %s\n' % (e.__class__.__name__, e))
        sys.exit(1)

    def _aliases(self, config):
        result = list(config['ROOT'].get('aliases', {}).items())
        global_config = load_global_config_parser()
        if global_config.has_section('alias'):
            result.extend(global_config.items('alias'))
        return result

    def _handle_arg_complete(self, config, command_map):
        words = os.environ['COMP_LINE'].split()
        command_name = words[1]

        choices = [x for x in command_map.keys()]
        for key, val in self._aliases(config):
            choices.append(key)

        if command_name not in choices:
            parser = ArgumentParser()
            parser.add_argument('command', choices=choices)
            argcomplete.autocomplete(parser)

        os.environ['COMP_LINE'] = ' '.join(words[:1] + words[2:])
        os.environ['COMP_POINT'] = str(
            int(os.environ['COMP_POINT']) - (len(command_name) + 1))

        return command_name

    def _find_alias(self, config, command_name):
        for key, val in self._aliases(config):
            if key == command_name:
                return val.split()
        return None

    def execute(self):
        """
        Execute command.

        Given the command-line arguments, this figures out which command is
        being run, creates a parser appropriate to that command, and runs it.
        """
        try:
            config = ConfigLoader().load()
            command_map = get_command_map(get_command_path(config))
        except Exception as e:
            self._handle_exception(e)

        if "_ARGCOMPLETE" in os.environ:
            command_name = self._handle_arg_complete(config, command_map)
        else:
            try:
                command_name = sys.argv[1]
            except IndexError:
                command_name = 'help'  # Display help if no arguments were given.

        if command_name == '--version':
            sys.stdout.write(get_version() + '\n')
        elif command_name == 'help':
            sys.stdout.write(
                self.main_help_text(command_map,
                                    commands_only=('--commands' in sys.argv)) +
                '\n')
        else:
            alias = self._find_alias(config, command_name)
            if alias and alias[0] in command_map:
                command_name = alias[0]
                sys.argv = sys.argv[:1] + alias + sys.argv[2:]

            if command_name not in command_map:
                print("Unknown dodo command: %s" % command_name)
                sys.exit(1)

            try:
                execute_script(command_map, command_name)
            except KeyboardInterrupt:
                print('\n')
                sys.exit(1)
            except Exception as e:
                self._handle_exception(e)


def execute_from_command_line(argv):
    """A simple method that runs a ManagementUtility."""
    utility = ManagementUtility()
    utility.execute()



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework.config import (get_command_path, Paths,
                                            projects_dir, ConfigLoader)
import glob
import os
import ruamel.yaml
import sys


def _args():  # noqa
    parser = ArgumentParser(
        description=('Print the value of /ROOT/<what>_dir. For example: ' +
                     '"dodo which src" prints the value of /ROOT/src_dir.'))
    parser.add_argument(
        'what',
        nargs='?',
    )

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--config',
                       action="store_true",
                       help='Print where the config file is')
    group.add_argument('--global-config',
                       action="store_true",
                       help='Print where the global config file is')
    group.add_argument(
        '--script',
        help='Print where the dodo command script with given name is')
    group.add_argument(
        '--dir',
        dest='directory',
        help=
        'Finds the X_dir value in the configuration, where X is the given option value'
    )
    group.add_argument('--decorators',
                       action="store_true",
                       help='Prints which command decorators are available')
    group.add_argument('--projects',
                       action="store_true",
                       help='Prints which projects are available')
    group.add_argument('--project',
                       action="store_true",
                       help='Prints name of the active project')
    group.add_argument('--layers',
                       action="store_true",
                       help='Prints which layers are active')
    group.add_argument(
        '--default-commands',
        action="store_true",
        help='Prints the directory where default commands are stored')
    group.add_argument(
        '--global-commands',
        action="store_true",
        help='Prints the directory where global commands are stored')

    return Dodo.parse_args(parser)


def _which_script(script):
    command_path = get_command_path(Dodo.get_config())
    for item in command_path.items:
        script_path = os.path.join(item, script + ".py")
        if os.path.exists(script_path):
            return os.path.realpath(script_path)

    for item in command_path.items:
        for yml in glob.glob(os.path.join(item, "dodo.*.yaml")):
            with open(yml) as ifs:
                if script in ruamel.yaml.round_trip_load(ifs.read()).keys():
                    return os.path.realpath(yml)

    return None


def _which_dir(directory):
    return Dodo.get_config("/ROOT/%s_dir" % directory, None)


if Dodo.is_main(__name__):
    args = _args()

    if args.config:
        if os.path.exists(Paths().project_dir()):
            print(os.path.join(Paths().res_dir(), "config.yaml"))
    elif args.global_config:
        sys.stdout.write(Paths().global_config_filename() + '\n')
    elif args.global_commands:
        sys.stdout.write(Paths().global_commands_dir() + '\n')
    elif args.default_commands:
        sys.stdout.write(Paths().default_commands_dir() + '\n')
    elif args.script:
        sys.stdout.write(_which_script(args.script) + '\n')
    elif args.directory:
        sys.stdout.write(_which_dir(args.directory) + '\n')
    elif args.decorators:
        sys.stdout.write(", ".join(sorted(Dodo._all_decorators().keys())) +
                         '\n')
    elif args.projects:
        projects = [
            x for x in os.listdir(projects_dir()) if os.path.isdir(
                os.path.join(projects_dir(), x, "dodo_commands", "res"))
        ]
        sys.stdout.write('\n'.join(sorted(projects)) + '\n')
    elif args.project:
        if Dodo.get_config("/ROOT/project_name", None):
            sys.stdout.write(Dodo.get_config("/ROOT/project_name") + '\n')
    elif args.layers:
        for layer_filename in ConfigLoader().get_layers(Dodo.get_config()):
            sys.stdout.write(layer_filename + '\n')
    elif args.what:
        x = _which_script(args.what) or _which_dir(args.what)
        if x:
            sys.stdout.write(x + '\n')
    else:
        if Dodo.get_config("/ROOT/project_name", None):
            sys.stdout.write(Dodo.get_config("/ROOT/project_name") + '\n')



"""Module for working with the dodo configurations."""

from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_expander import ConfigExpander
from dodo_commands.framework.config_expander import Key, KeyNotFound  # noqa
from dodo_commands.framework.util import symlink
from plumbum import local
from six.moves import configparser
import glob
from fnmatch import fnmatch
import hashlib
import json
import os
import argparse
from os.path import dirname
import re
import ruamel.yaml
import sys

try:
    from dotenv import dotenv_values
except ImportError:
    # Stub
    def dotenv_values(x):
        raise CommandError(
            "Package python-dotenv not installed. Please run: pip install python-dotenv."
        )


def _is_windows():
    return os.name == 'nt'


def _ext():
    return '.exe' if _is_windows() else ''


class Paths:
    # Cached result of finding the current project dir
    _project_dir = None

    def __init__(self, project_dir=None):
        self._project_dir = project_dir or Paths._project_dir
        if not self._project_dir:
            bin_dir = os.path.realpath(dirname(sys.executable))
            marker_file = os.path.join(bin_dir, '.dodo_project_dir_marker')
            Paths._project_dir = (dirname(dirname(dirname(bin_dir)))
                                  if os.path.exists(marker_file) else "")
            self._project_dir = Paths._project_dir

    def home_dir(self, expanduser=True):
        return os.path.expanduser('~') if expanduser else '~'

    def global_config_dir(self, expanduser=True):
        return os.path.join(self.home_dir(expanduser), '.dodo_commands')

    def global_config_filename(self):
        return os.path.join(self.global_config_dir(), 'config')

    def default_commands_dir(self, expanduser=True):
        return os.path.join(self.global_config_dir(expanduser),
                            'default_commands')

    def global_commands_dir(self, expanduser=True):
        return os.path.join(self.global_config_dir(expanduser), 'commands')

    def virtual_env_dir(self):
        return os.path.join(self.project_dir(), 'dodo_commands', 'env')

    def virtual_env_bin_dir(self):
        return os.path.join(self.virtual_env_dir(),
                            'Scripts' if _is_windows() else 'bin')

    def pip(self):
        return os.path.join(self.virtual_env_bin_dir(), 'pip' + _ext())

    def site_packages_dir(self):
        python = local[os.path.join(self.virtual_env_bin_dir(),
                                    "python" + _ext())]

        result = python(
            "-c", "from distutils.sysconfig import get_python_lib; " +
            "print(get_python_lib())")
        while result[-1] in ['\n', '\r']:
            result = result[:-1]
        return result

    def project_dir(self):
        """Return the root dir of the current project."""
        return self._project_dir

    def res_dir(self):
        return os.path.join(self.project_dir(), "dodo_commands", "res")

    def package_dir(self):
        import dodo_commands
        return os.path.dirname(dodo_commands.__file__)

    def extra_dir(self):
        return os.path.join(self.package_dir(), "extra")


def load_global_config_parser():
    config_parser = configparser.ConfigParser()
    config_parser.read(Paths().global_config_filename())
    return config_parser


def write_global_config_parser(config_parser):
    """Save configuration."""
    with open(Paths().global_config_filename(), "w") as f:
        config_parser.write(f)


def projects_dir():
    return os.path.expanduser(load_global_config_parser().get(
        "settings", "projects_dir"))


_global_config = """[settings]
projects_dir=~/projects
python_interpreter=python
diff_tool=diff

[recent]
"""


def _touch_init_py(dir_name):
    init_py = os.path.join(dir_name, "__init__.py")
    if not os.path.exists(init_py):
        with open(init_py, 'w'):
            pass


def create_global_config():
    """Create config file and default_commands dir."""
    base_dir = Paths().global_config_dir()
    if not os.path.exists(base_dir):
        os.mkdir(base_dir)

    config_filename = Paths().global_config_filename()
    if not os.path.exists(config_filename):
        with open(config_filename, 'w') as f:
            f.write(_global_config)

    default_commands_dir = Paths().default_commands_dir()
    if not os.path.exists(default_commands_dir):
        os.mkdir(default_commands_dir)
        _touch_init_py(default_commands_dir)
        symlink(
            os.path.join(Paths().extra_dir(), "dodo_standard_commands"),
            os.path.join(Paths().default_commands_dir(),
                         "dodo_standard_commands"))

    global_commands_dir = Paths().global_commands_dir()
    if not os.path.exists(global_commands_dir):
        os.mkdir(global_commands_dir)
        _touch_init_py(global_commands_dir)


def merge_into_config(config, layer, xpath=None):
    def _is_list(x):
        return isinstance(x, type(list()))

    def _is_dict(x):
        return isinstance(x, type(dict()))

    def _raise(xpath):
        raise CommandError("Cannot merge configurations. Check key /%s" %
                           '/'.join(new_xpath))

    xpath = xpath or []
    for key, val in (layer or {}).items():
        new_xpath = xpath + [key]

        if key not in config:
            config[key] = val
        elif _is_dict(val):
            if not _is_dict(config[key]):
                _raise(new_xpath)
            merge_into_config(config[key], val, new_xpath)
        elif _is_list(val):
            if not _is_list(config[key]):
                _raise(new_xpath)
            config[key].extend(val)
        else:
            config[key] = val


class ConfigIO:
    """Read and write the dodo configuration."""

    def __init__(self, config_base_dir=None):
        """Arg config_base_dir is where config files are searched.

        Arg config_base_dir defaults to Paths().res_dir().
        """
        self.config_base_dir = (Paths().res_dir() if config_base_dir is None
                                else config_base_dir)

    def _path_to(self, post_fix_paths):
        """Return path composed of config_base_dir and post_fix_paths list."""
        return os.path.join(self.config_base_dir, *post_fix_paths)

    def glob(self, patterns):
        """Returns list of filenames"""

        def add_prefix(filename):
            return (filename
                    if filename.startswith('/') else self._path_to([filename]))

        patterns = [
            add_prefix(os.path.expanduser(pattern)) for pattern in patterns
        ]

        result = []
        for pattern in patterns:
            result.extend([x for x in glob.glob(pattern)])
        return result

    def load(self, config_filename='config.yaml'):
        """Get configuration."""
        full_config_filename = self._path_to([config_filename])
        if not os.path.exists(full_config_filename):
            return None

        with open(full_config_filename) as f:
            config = ruamel.yaml.round_trip_load(f.read())
        return config

    def save(self, config, config_filename='config.yaml'):
        """Write config to config_filename."""
        with open(self._path_to([config_filename]), 'w') as f:
            return f.write(ruamel.yaml.round_trip_dump(config))


class ConfigLoader:
    """Load the project's dodo config and expand it."""

    def __init__(self, config_io=None):
        self.config_io = config_io or ConfigIO()

    def _add_to_config(self, config, section, key, value):
        if section in config:
            if key not in config[section]:
                config[section][key] = value

    def _system_commands_dir(self):
        """Return directory where system command scripts are stored"""
        import dodo_commands.dodo_system_commands
        return os.path.dirname(dodo_commands.dodo_system_commands.__file__)

    def _extend_config(self, config):
        """Add special values to the project's config"""
        project_dir = Paths().project_dir()
        if project_dir:
            self._add_to_config(config, "ROOT", "project_name",
                                os.path.basename(project_dir))
            self._add_to_config(config, "ROOT", "project_dir", project_dir)
            self._add_to_config(config, "ROOT", "res_dir", Paths().res_dir())

    def _extend_command_path(self, config):
        """Add the system commands to the command path"""
        self._add_to_config(config, "ROOT", "command_path", [])
        config['ROOT']['command_path'].append(self._system_commands_dir())
        if not Paths().project_dir():
            config['ROOT']['command_path'].append(
                os.path.join(Paths().default_commands_dir(), '*'))

    def _report(self, x):
        sys.stderr.write(x)
        sys.stderr.flush()

    def get_layers(self, config):
        layer_filenames = []
        for pattern in config.get('ROOT', {}).get('layers', []):
            filenames = self.config_io.glob([pattern])
            if not filenames:
                print("Warning, no layers found for pattern: %s" % pattern)
            layer_filenames.extend(filenames)

        parser = argparse.ArgumentParser()
        parser.add_argument('--layer', action='append', help=argparse.SUPPRESS)
        args, _ = parser.parse_known_args(
            [x for x in sys.argv if x not in ('--help', '-h')])
        extra_layers = args.layer or []

        for extra_layer_filename in self.config_io.glob(extra_layers):
            # Remove layers in the same group, because by definition
            # we should not use both foo.bar.yaml and foo.baz.yaml.
            parts = os.path.basename(extra_layer_filename).split('.')
            if len(parts) == 3:
                pattern = os.path.join(os.path.dirname(extra_layer_filename),
                                       parts[0] + '.*.*')
                layer_filenames = [
                    x for x in layer_filenames if not fnmatch(x, pattern)
                ]

            if extra_layer_filename not in layer_filenames:
                layer_filenames.append(extra_layer_filename)

        return layer_filenames

    def load(self, extend=True):
        fallback_config = dict(ROOT={})
        try:
            config = self.config_io.load() or fallback_config
            for layer_filename in self.get_layers(config):
                layer = self.config_io.load(layer_filename)
                merge_into_config(config, layer)

        except ruamel.yaml.scanner.ScannerError:
            config = fallback_config
            self._report(
                "There was an error while loading the configuration. "
                "Run 'dodo diff' to compare your configuration to the "
                "default one.\n")

        if extend:
            self._extend_command_path(config)
            self._extend_config(config)

        extra_vars = dict()

        def _load_env(dotenv_file):
            if not os.path.exists(dotenv_file):
                self._report("Dotenv file not found: %s\n" % dotenv_file)
            extra_vars.update(dotenv_values(dotenv_file))

        # Call dotenv_values for every item of /ENV/dotenv
        callbacks = {}
        for idx, _ in enumerate(config['ROOT'].get('dotenv_files', [])):
            callbacks['/ROOT/dotenv_files/%d' % idx] = _load_env

        ConfigExpander(extra_vars).run(config, callbacks=callbacks)
        return config


class CommandPath:
    """Read search paths for command scripts from the configuration."""

    def __init__(self, include_patterns, exclude_patterns):
        excluded_command_dirs = self._create_items(exclude_patterns)

        self.items = sorted([
            x for x in self._create_items(include_patterns)
            if x not in excluded_command_dirs
        ])

        basenames = [os.path.basename(x) for x in self.items]
        for basename in basenames:
            if basenames.count(basename) > 1:
                raise CommandError("More than 1 command path with name %s" %
                                   basename)

    def _create_items(self, patterns):
        result = []
        for pattern in patterns:
            for x in glob.glob(os.path.expanduser(pattern)):
                if os.path.isdir(
                        x) and not os.path.basename(x).startswith('_'):
                    result.append(x)
        return result

    def _create_search_path_dir(self):
        hash_code = hashlib.md5(json.dumps(
            self.items).encode('utf-8')).hexdigest()
        search_path_dir = os.path.join(Paths().global_config_dir(),
                                       "search_path", hash_code)

        if not os.path.exists(search_path_dir):
            os.makedirs(search_path_dir)
            open(os.path.join(search_path_dir, "__init__.py"), 'a').close()
            for item in self.items:
                basename = os.path.basename(item)
                target_path = os.path.join(search_path_dir, basename)
                if not os.path.exists(target_path):
                    symlink(item, target_path)

        return search_path_dir

    def extend_sys_path(self):
        search_path_dir = self._create_search_path_dir()
        if search_path_dir not in sys.path:
            sys.path.append(search_path_dir)


def expand_keys(config, text):
    result = ""
    val_terms = re.split('\$\{([^\}]+)\}', text)
    for idx, term in enumerate(val_terms):
        if idx % 2:
            xpath = [k for k in term.split("/") if k]
            str_rep = json.dumps(Key(config, xpath).get())
            if str_rep.startswith('"') and str_rep.endswith('"'):
                str_rep = str_rep[1:-1]
            result += str_rep
        else:
            result += term
    return result


def get_command_path(config=None):
    config = config or ConfigLoader().load()
    root_config = config.get('ROOT', {})
    command_path = root_config.get('command_path', [])
    command_path_exclude = root_config.get('command_path_exclude', [])
    return CommandPath(include_patterns=command_path,
                       exclude_patterns=command_path_exclude)



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.config import (load_global_config_parser,
                                            ConfigIO, Paths, ConfigLoader)
from dodo_commands.framework.config_expander import Key
from six.moves import configparser
import os
import sys


def _args():
    parser = ArgumentParser(description=('Edit the dodo configuration files'))

    parser.add_argument(
        '--key', help='Only edit this key. Used in conjunction with --val')
    parser.add_argument(
        '--val',
        help='The value to be used in conjunction with the --key option')

    args = Dodo.parse_args(parser)

    if args.key or args.val:
        if not args.key and args.val:
            raise CommandError(
                "The options --key and --val should always be used together")

    args.editor = load_global_config_parser().get("settings", "config_editor")
    args.res_dir = Dodo.get_config('/ROOT/res_dir')
    return args


if Dodo.is_main(__name__, safe=('--key' not in sys.argv)):
    try:
        args = _args()
    except configparser.NoOptionError as e:
        raise CommandError("{error}. Please check {filename}".format(
            error=str(e), filename=Paths().global_config_filename()))

    if args.key and args.val:
        config = ConfigIO().load()
        xpath = [x for x in args.key.split('/') if x]
        key = Key(config, xpath)
        key.set(args.val)
        ConfigIO().save(config)
        sys.exit(0)

    yaml_filenames = ([os.path.join(args.res_dir, 'config.yaml')] +
                      ConfigLoader().get_layers(Dodo.get_config()))

    yaml_filenames.append(Paths().global_config_filename())
    Dodo.run(args.editor.split() + yaml_filenames, cwd='.')



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
from dodo_commands.framework.util import bash_cmd


def _args():
    parser = ArgumentParser()
    parser.add_argument('--name')
    args = Dodo.parse_args(parser)
    args.name = args.name or 'dodo'
    return args


if Dodo.is_main(__name__, safe=True):
    args = _args()

    tmp_filename = '/tmp/%s_autocomplete.sh' % args.name
    Dodo.run(
        bash_cmd(
            'register-python-argcomplete %s > %s' % (args.name, tmp_filename)))
    Dodo.run([
        'sudo', 'mv', tmp_filename,
        '/etc/bash_completion.d/%s_autocomplete.sh' % args.name
    ])



from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.config import projects_dir, Paths
from dodo_commands.framework.util import is_using_system_dodo
import os


def _args():
    parser = ArgumentParser()
    parser.add_argument("project_name")
    parser.add_argument("new_project_name")
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if not is_using_system_dodo():
        raise CommandError(
            "Please deactivate your dodo project first by running 'deactivate'."
        )

    old_project_dir = os.path.join(projects_dir(), args.project_name)
    if not os.path.exists(old_project_dir):
        raise CommandError("Project dir not found: %s" % old_project_dir)
    old_paths = Paths(old_project_dir)

    new_project_dir = os.path.join(projects_dir(), args.new_project_name)
    if os.path.exists(new_project_dir):
        raise CommandError(
            "New project dir already exists: %s" % new_project_dir)
    new_paths = Paths(new_project_dir)

    Dodo.run(['mv', old_project_dir, new_project_dir], cwd='.')

    def fn(x):
        return os.path.join(new_paths.virtual_env_bin_dir(), x)

    replace_project_dir = 's#%s#%s#g' % (old_project_dir, new_project_dir)
    for script in (fn('dodo'), fn('activate')):
        Dodo.run(['sed', '-i', '-e', replace_project_dir, script], cwd='.')

    replace_prompt = 's#PS1="(%s) $PS1\"#PS1="(%s) $PS1"#g' % (
        args.project_name, args.new_project_name)
    Dodo.run(['sed', '-i', '-e', replace_prompt, fn('activate')], cwd='.')



import os
import sys
import ruamel.yaml
import subprocess
import traceback
from importlib import import_module
from dodo_commands.framework.config import Paths
from dodo_commands.framework.command_map import CommandMapItem
from dodo_commands.framework.util import query_yes_no


class PythonCommandMapItem(CommandMapItem):
    def __init__(self, group, filename):
        super(PythonCommandMapItem, self).__init__(group, filename, 'py')
        self.package_path = group


class PythonCommandHandler:
    def add_commands_to_map(self, command_path, file_map, command_map):
        command_path.extend_sys_path()
        for command_dir, files in file_map.items():
            for file in files:
                command_name, ext = os.path.splitext(os.path.basename(file))
                if ext == '.py' and not command_name.startswith('_'):
                    command_map[command_name] = PythonCommandMapItem(
                        group=os.path.basename(command_dir), filename=file)

    def execute(self, command_map_item, command_name):
        def install_packages(meta_data_filename):
            """Pip install packages found in meta_data_filename."""
            with open(meta_data_filename) as f:
                meta_data = ruamel.yaml.round_trip_load(f.read())
                print("This command wants to install additional packages:\n")
                print(meta_data['requirements'])
                if query_yes_no("Install (yes), or abort (no)?"):
                    print("\n--- Installing from %s ---" % meta_data_filename)
                    pip = Paths().pip()
                    subprocess.check_call([pip, "install"] +
                                          meta_data['requirements'])
                    print("--- Done ---\n\n")
                else:
                    sys.exit(1)

        import_path = '%s.%s' % (command_map_item.package_path, command_name)
        if command_map_item.package_path in ("", None, "."):
            import_path = command_name

        try:
            import_module(import_path)
        except ImportError as e:
            try:
                base_path = import_module(
                    command_map_item.package_path).__path__[0]
                meta_data_filename = os.path.join(base_path,
                                                  command_name + ".meta")
                if os.path.exists(meta_data_filename):
                    install_packages(meta_data_filename)
                    import_module(import_path)
                else:
                    traceback.print_exc(e)
                    sys.exit(1)
            except ImportError as e:
                traceback.print_exc(e)
                sys.exit(1)



import ruamel.yaml
import re
from argparse import ArgumentParser
from dodo_commands.framework import Dodo


def _args():
    parser = ArgumentParser(description="Print the full configuration.")
    parser.add_argument('key', nargs='?')
    args = Dodo.parse_args(parser)
    if args.key and args.key.startswith('/'):
        args.key = args.key[1:]
    return args


if Dodo.is_main(__name__):
    args = _args()
    if args.key:
        contents = Dodo.get_config(args.key)
        if isinstance(contents, str):
            print(contents)
        else:
            print("%s" % ruamel.yaml.round_trip_dump(contents))
    else:
        content = re.sub(
            r'^([0-9_A-Z]+\:)$',
            r'\n\1',
            ruamel.yaml.round_trip_dump(Dodo.get_config()),
            flags=re.MULTILINE)
        print(re.sub(r'^\n\n', r'\n', content, flags=re.MULTILINE))



import sys
from argparse import ArgumentParser
from dodo_commands.framework import Dodo, CommandError
from dodo_commands.framework.util import filter_choices, query_yes_no
from plumbum.cmd import docker
from six.moves import input as raw_input


def _container_type_names():
    return list(Dodo.get_config('/DOCKER/container_types', {}).keys())


def _args():
    parser = ArgumentParser()
    parser.add_argument('container_type',
                        choices=_container_type_names(),
                        nargs='?')
    parser.add_argument('name', nargs='?')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('--interactive', '-i', action='store_true')
    group.add_argument('--replace',
                       action='store_true',
                       help='Replace existing container')

    args = Dodo.parse_args(parser)
    args.container_types = Dodo.get_config("/DOCKER/container_types")
    args.project_name = Dodo.get_config('/ROOT/project_name')

    return args


def _exists(name):
    return docker("ps", "-a", "--quiet", "--filter", "name=" + name)


def _create_container(container_types, container_type_name, name, replace):
    dirs = container_types.get(container_type_name, {})['dirs']
    image = container_types.get(container_type_name, {})['image']

    exists = _exists(name)

    if exists and replace:
        Dodo.run(['docker', 'rm', name])

    if not exists or replace:
        cmd_args = [
            "docker",
            "create",
            "--name",
            name,
        ]

        for path in dirs:
            cmd_args.extend(["-v", path])

        cmd_args.append(image)
        Dodo.run(cmd_args)

    Dodo.runcmd([
        'dodo', 'edit-config',
        '--key=/DOCKER/containers/%s' % container_type_name,
        '--val=%s' % name
    ])


if Dodo.is_main(__name__, safe=True):
    args = _args()

    if args.interactive:
        container_type_names = _container_type_names()
        for idx, container_type in enumerate(container_type_names):
            print('%d - %s' % (idx + 1, container_type))

        raw_choice = raw_input("Select a container type: ")
        selected_container_type_names, span = filter_choices(
            container_type_names, raw_choice)
        if span == [0, len(raw_choice)]:
            for container_type_name in selected_container_type_names:
                default_container_name = 'dc_%s_%s' % (args.project_name,
                                                       container_type_name)
                name = raw_input(
                    "Enter a name for container type %s [%s]: " %
                    (container_type_name,
                     default_container_name)) or default_container_name

                if _exists(name):
                    replace = query_yes_no(
                        'A container with this name exists. ' +
                        'To replace choose yes. To quit choose no. Replace it?'
                    )
                    if not replace:
                        sys.exit(0)

                _create_container(args.container_types,
                                  container_type_name,
                                  name,
                                  replace=True)
        else:
            raise CommandError("Syntax error")

        sys.exit(0)
    else:
        if not args.container_type:
            raise CommandError('Argument container_type is required')
        if not args.name:
            raise CommandError('Argument name is required')
        _create_container(args.container_types, args.container_type, args.name,
                          args.replace)



from dodo_commands.framework import CommandError, Dodo
from dodo_standard_commands.decorators.docker import (Decorator as
                                                      DockerDecorator)
from dodo_commands.framework.config import expand_keys
from dodo_commands.framework.util import indent
import os
import plumbum
import ruamel.yaml

filters = {}
tests = {}


def register_diagnose_filter(name=None):
    def decorator(func):
        global filters
        filters[name or func.__name__] = func
        return func

    return decorator


def register_diagnose_test(name=None):
    def decorator(func):
        global tests
        tests[name or func.__name__] = func
        return func

    return decorator


@register_diagnose_filter('dodo_expand')
def _dodo_expand(key_str,
                 key=None,
                 layout=None,
                 quote_key=True,
                 quote_val=True,
                 link=None):
    def _quote_val(val):
        if link or not quote_val:
            return val
        return '`' + val + '`'

    def _quote_key(key):
        if link or not quote_key:
            return key
        return '`' + key + '`'

    if key is None:
        key = False

    if link is None:
        link = False

    if "${" not in key_str:
        val = Dodo.get_config(key_str)
        key_str = '${' + key_str + '}'
    else:
        val = expand_keys(Dodo.get_config(), key_str)

    quoted_key_str = _quote_key(key_str)
    val_str = _quote_val(str(val))

    if layout is None:
        layout = isinstance(val, type(dict()))

    result = ""
    if link:
        prefix = "" if os.path.exists(val_str) else "value:"
        result = "`%s <%s%s>`_" % (quoted_key_str, prefix, val_str)
    elif layout:
        if key:
            result += "%s:\n" % quoted_key_str
        result += ("\n\n.. code-block:: yaml\n\n" + indent(
            "# %s\n" % key_str, 4) + indent(
                ruamel.yaml.round_trip_dump(val, indent=4).strip(), amount=4))
    else:
        if key:
            result += "%s = " % quoted_key_str
        result += val_str

    return result


@register_diagnose_filter('leaf')
def _leaf(str_arg):
    return str_arg.split("/")[-1]


def _docker():
    try:
        return plumbum.local['docker']
    except plumbum.commands.processes.CommandNotFound:
        raise CommandError("Docker is not installed")


def _image(key_or_name):
    image = Dodo.get_config(key_or_name, None)
    if not image:
        _, image, _ = DockerDecorator.docker_node(Dodo.get_config, key_or_name,
                                                  "")
    return image


@register_diagnose_test('existing_docker_image')
def _is_existing_docker_image(key):
    image = _image(key)
    return _docker()("images", "-q", image)


@register_diagnose_test('existing_container')
def _is_existing_container(key):
    image = _image(key)
    return image in _docker()("ps", "-a", "--filter=name=%s" % image)


@register_diagnose_test('path_exists')
def _path_exists(path):
    return os.path.exists(path)



"""Decorates command lines with docker arguments."""
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.config import merge_into_config
from dodo_commands.framework import Dodo, CommandError
from fnmatch import fnmatch
from plumbum import local
import os


def _is_tuple(x):
    return hasattr(x, "__len__") and not isinstance(x, type(str()))


class Decorator:  # noqa
    @classmethod
    def _add_docker_volume_list(cls, docker_config, root_node):
        volume_list = docker_config.get('volume_list', [])
        volume_map = docker_config.get('volume_map', {})

        volume_map_strict = docker_config.get('volume_map_strict', {})
        for key, val in volume_map_strict.items():
            if not os.path.exists(key):
                raise CommandError(
                    "Path in volume_map_strict not found: %s" % key)
            volume_map[key] = val

        options = (
            ['--volume=%s:%s' % (x, x) for x in volume_list] +
            ['--volume=%s:%s' % key_val for key_val in volume_map.items()])
        for x in options:
            root_node['volume'].append(x)

    @classmethod
    def _add_docker_publish_list(cls, docker_config, root_node):
        publish_list = docker_config.get('publish_list', [])
        publish_map = docker_config.get('publish_map', {})
        options = (
            ['--publish=%s:%s' % (x, x) for x in publish_list] +
            ['--publish=%s:%s' % key_val for key_val in publish_map.items()])
        for x in options:
            root_node['publish'].append(x)

    @classmethod
    def _add_docker_volumes_from_list(cls, docker_config, root_node):
        volumes_from_list = docker_config.get('volumes_from_list', [])
        options = (['--volumes-from=%s' % x for x in volumes_from_list])
        for x in options:
            root_node['volumes-from'].append(x)

    @classmethod
    def _add_docker_variable_list(cls, docker_config, root_node):
        variable_list = docker_config.get('variable_list', [])
        variable_map = docker_config.get('variable_map', {})
        options = (
            ['--env=%s' % x for x in variable_list
             ] + ['--env=%s=%s' % key_val for key_val in variable_map.items()])
        for x in options:
            root_node['env'].append(x)

    @classmethod
    def _add_linked_container_list(cls, docker_config, root_node):
        options = ['--link=%s' % x for x in docker_config.get('link_list', [])]
        for x in options:
            root_node['link'].append(x)

    @classmethod
    def merged_options(cls, get_config, command_name):
        merged = {}
        for patterns, docker_config in get_config('/DOCKER/options',
                                                  {}).items():
            should_merge = False
            for pattern in (patterns if _is_tuple(patterns) else [patterns]):
                if pattern.startswith('!'):
                    should_merge = should_merge and not fnmatch(
                        command_name, pattern[1:])
                else:
                    should_merge = should_merge or fnmatch(
                        command_name, pattern)
            if should_merge:
                merge_into_config(merged, docker_config)
        return merged

    @classmethod
    def docker_node(cls, get_config, command_name, cwd):
        merged = cls.merged_options(get_config, command_name)

        docker_node = ArgsTreeNode("docker", args=['docker', 'run'])
        docker_node.add_child(ArgsTreeNode("name"))
        docker_node.add_child(ArgsTreeNode("basic"))
        docker_node.add_child(ArgsTreeNode("link"))
        docker_node.add_child(ArgsTreeNode("publish"))
        docker_node.add_child(ArgsTreeNode("volume", is_horizontal=False))
        docker_node.add_child(ArgsTreeNode("volumes-from"))
        docker_node.add_child(ArgsTreeNode("env", is_horizontal=False))
        docker_node.add_child(ArgsTreeNode("other"))
        docker_node.add_child(ArgsTreeNode("workdir"))
        docker_node.add_child(ArgsTreeNode("image"))

        cls._add_linked_container_list(merged, docker_node)
        cls._add_docker_publish_list(merged, docker_node)
        cls._add_docker_volume_list(merged, docker_node)
        cls._add_docker_volumes_from_list(merged, docker_node)
        cls._add_docker_variable_list(merged, docker_node)

        if merged.get('rm', True):
            docker_node['basic'].append('--rm')

        if merged.get('is_interactive', True):
            docker_node['basic'].append('--interactive')
            docker_node['basic'].append('--tty')

        if merged.get('is_daemon', False):
            docker_node['basic'].append('-d')

        name = merged.get('name', command_name)
        if name:
            docker_node["name"].append("--name=%s" % name)

        for key_val in get_config('/ENVIRONMENT/variable_map', {}).items():
            docker_node['env'].append("--env=%s=%s" % key_val)

        for x in merged.get('extra_options', []):
            docker_node['other'].append(x)

        if cwd:
            docker_node['workdir'].append('--workdir=' + cwd)

        image = merged.get('image')
        if not image:
            raise CommandError(
                "No docker image found in /DOCKER/options for command %s" %
                command_name)
        docker_node['image'].append(image)

        return docker_node, image, name

    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            '--kill-existing',
            action='store_true',
            help="Kill and remove existing docker container with the same name"
        )

    def _kill_existing_container(self, name):
        docker = local['docker']
        live_containers = docker("ps", "--format={{.Names}}",
                                 "--filter=name=^/%s$" % name)
        if len(live_containers):
            docker("stop", name)

        exited_containers = docker("ps", "-a", "--format={{.Names}}",
                                   "--filter=name=^/%s$" % name)
        if len(exited_containers):
            docker("rm", name)

    def modify_args(self, dodo_args, root_node, cwd):  # noqa
        docker_node, _, docker_name = self.docker_node(Dodo.get_config,
                                                       Dodo.command_name, cwd)

        if getattr(dodo_args, 'kill_existing', False):
            self._kill_existing_container(docker_name)

        docker_node.add_child(root_node)
        return docker_node, local.cwd



from argparse import ArgumentParser
from dodo_commands.framework import Dodo
import os


def _args():
    parser = ArgumentParser()
    parser.add_argument(
        'container_type', choices=Dodo.get_config('/DOCKER/containers').keys())
    parser.add_argument('output_dir')
    parser.add_argument('--reverse', action='store_true')
    args = Dodo.parse_args(parser)
    return args


def _docker_run(output_dir, args):
    Dodo.run(
        [
            'docker', 'run', '--rm', '--volumes-from', container_name,
            '--volume',
            '%s:/tmp/docker-snapshot' % output_dir, container_type['image']
        ] + args,
        cwd='.')


if Dodo.is_main(__name__, safe=True):
    args = _args()
    container_name = Dodo.get_config('/DOCKER/containers/' +
                                     args.container_type)

    Dodo.get_config('/DOCKER')['options'].setdefault('docker-snapshot', {})
    Dodo.get_config('/DOCKER')['options']['docker-snapshot'][
        'volumes_from'] = container_name

    container_type = Dodo.get_config('/DOCKER/container_types/' +
                                     args.container_type)
    for path in container_type['dirs']:
        host_output_path = os.path.join(args.output_dir, path[1:])
        docker_output_path = os.path.join('/tmp/docker-snapshot', path[1:])

        if args.reverse:
            _docker_run(
                args.output_dir,
                ['/bin/bash', '-c', 'rm -rf ' + os.path.join(path, '*')])
            _docker_run(args.output_dir, [
                '/bin/bash', '-c',
                'cp -rf %s %s' % (os.path.join(docker_output_path, '*'), path)
            ])
        else:
            if os.path.exists(host_output_path):
                Dodo.run(['rm', '-rf', host_output_path], cwd='.')
            Dodo.run(['mkdir', '-p',
                      os.path.dirname(host_output_path)],
                     cwd='.')
            _docker_run(
                args.output_dir,
                ['cp', '-rf', path,
                 os.path.dirname(docker_output_path)])



from dodo_commands.framework.util import query_yes_no
from importlib import import_module
import argcomplete
import argparse
import fnmatch
import os
import sys
from dodo_commands.framework.config import (ConfigLoader, get_command_path)
from dodo_commands.framework.config_expander import Key, KeyNotFound
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.command_error import CommandError
from plumbum import FG, ProcessExecutionError, local


def _ask_confirmation(args, cwd, is_echo, is_confirm):
    """Ask the user whether to continue with executing func."""

    def to_str():
        return args.to_str(slash=args.children)

    if is_echo:
        print(to_str())
        return False

    if is_confirm:
        print("(%s) %s" % (cwd, to_str()))
        if not query_yes_no("confirm?"):
            return False
        else:
            print("")

    return True


class DecoratorScope:
    def __init__(self, decorator_name):
        self.decorators = Dodo.get_config('/ROOT').setdefault(
            'decorators', {}).setdefault(decorator_name, [])

    def __enter__(self):  # noqa
        self.decorators.append(Dodo.command_name)

    def __exit__(self, type, value, traceback):  # noqa
        self.decorators.remove(Dodo.command_name)


class ConfigArg:
    def __init__(self, config_key, *args, **kwargs):
        self.config_key = config_key
        self.args = args
        self.kwargs = kwargs

    @property
    def arg_name(self):
        return self.args[0].strip('-').replace('-', '_')

    @property
    def xpath(self):
        return [x for x in self.config_key.split('/') if x]


class Dodo:
    package_path = None
    command_name = None
    safe = True

    _args = argparse.Namespace()
    _config = None

    @classmethod
    def get_config(cls, key='', default_value="__not_set_234234__"):  # noqa
        if cls._config is None:
            cls._config = ConfigLoader().load()

        try:
            xpath = [k for k in key.split("/") if k]
            return Key(cls._config, xpath).get()
        except KeyNotFound:
            if default_value == "__not_set_234234__":
                raise
            return default_value

    @classmethod
    def is_main(cls, name, safe=None):
        if safe is not None:
            cls.safe = safe

        if name == '__main__':
            import __main__ as main
            cls.command_name = os.path.basename(main.__file__)
            return True

        return (cls.command_name and cls.package_path
                and name == cls.package_path + '.' + cls.command_name)

    @classmethod
    def _get_decorators(cls):
        return [
            cls._load_decorator(name, directory)
            for name, directory in cls._all_decorators().items()
            if cls._uses_decorator(name)
        ]

    @classmethod
    def _uses_decorator(cls, decorator_name):
        patterns = (cls.get_config().get('ROOT',
                                         {}).get('decorators',
                                                 {}).get(decorator_name, []))
        command_name = cls.command_name
        approved = [
            pattern for pattern in patterns if not pattern.startswith("!")
            and fnmatch.filter([command_name], pattern)
        ]
        rejected = [
            pattern for pattern in patterns if pattern.startswith("!")
            and fnmatch.filter([command_name], pattern[1:])
        ]
        return len(approved) and not len(rejected)

    @classmethod
    def _load_decorator(cls, name, directory):
        """Load and return decorator class in module with given name."""
        return import_module(directory + "." + name).Decorator()

    @classmethod
    def _all_decorators(cls):
        """Returns a mapping from decorator name to its directory."""
        command_path = get_command_path(cls.get_config())
        command_path.extend_sys_path()
        result = {}
        for item in command_path.items:
            try:
                module_path = os.path.basename(item) + ".decorators"
                module = import_module(module_path)
                for decorator in os.listdir(module.__path__[0]):
                    name, ext = os.path.splitext(decorator)
                    if ext == '.py' and name != '__init__':
                        result[name] = module_path
            except ImportError:
                continue
        return result

    @classmethod
    def _add_config_args(cls, parser, config_args=None):
        show_help = '--help' in sys.argv

        for config_arg in (config_args or []):
            key = Key(cls.get_config(), config_arg.xpath)
            key_exists = key.exists()

            if show_help or not key_exists:
                kwargs = dict(config_arg.kwargs)
                help_text = kwargs.get('help') or ''
                sep = '. ' if help_text else ''
                if key_exists:
                    value = str(key.get())
                    formatted_value = value[:50] + (value[50:] and '...')
                    extra_help = ('Read from config: %s = %s.' %
                                  (config_arg.config_key, formatted_value))
                else:
                    extra_help = ('Configuration key is %s' %
                                  config_arg.config_key)

                kwargs['help'] = help_text + sep + extra_help
                parser.add_argument(*config_arg.args, **kwargs)

    @classmethod
    def parse_args(cls, parser, config_args=None):
        parser.add_argument('--traceback',
                            action='store_true',
                            help=argparse.SUPPRESS)

        parser.add_argument(
            '-c',
            '--confirm',
            action='store_true',
            help="Confirm each performed action before its execution")

        parser.add_argument('-e',
                            '--echo',
                            action='store_true',
                            help="Print all commands instead of running them")

        # The --layer option is handled by the argument parser in ConfigLoader.
        # We need to add it to this parser as well, otherwise it will complain.
        # Yes, quite hacky...
        parser.add_argument('--layer', action='append', help=argparse.SUPPRESS)

        for decorator in cls._get_decorators():
            decorator.add_arguments(parser)

        cls._add_config_args(parser, config_args)
        argcomplete.autocomplete(parser)
        if os.path.splitext(sys.argv[0])[1].lower() == '.py':
            parser.prog = sys.argv[0]
            first_arg_index = 1
        else:
            first_arg_index = 2
            parser.prog = "%s %s" % (os.path.basename(
                sys.argv[0]), sys.argv[1])
        cls._args = parser.parse_args(sys.argv[first_arg_index:])

        if config_args:
            for config_arg in config_args:
                key = Key(cls.get_config(), config_arg.xpath)
                if key.exists():
                    setattr(cls._args, config_arg.arg_name, key.get())

        if cls._args.echo and not cls.safe:
            raise CommandError(
                "The --echo option is not supported for unsafe commands.\n" +
                "Since this command is marked as unsafe, some operations will "
                + "be performed directly, instead of echoed to the console. " +
                "Use --confirm if you wish to execute with visual feedback. ")

        if cls._args.confirm and not cls.safe:
            if not query_yes_no(
                    "Since this command is marked as unsafe, some operations will "
                    +
                    "be performed without asking for confirmation. Continue?",
                    default="no"):
                sys.exit(1)

        return cls._args

    @classmethod
    def run(cls, args, cwd=None, quiet=False, capture=False):
        root_node = ArgsTreeNode('original_args', args=args)
        for decorator in cls._get_decorators():
            root_node, cwd = decorator.modify_args(cls._args, root_node, cwd)

        if not _ask_confirmation(root_node, cwd or local.cwd,
                                 getattr(cls._args, 'echo', False),
                                 getattr(cls._args, 'confirm', False)):
            return False

        if cwd and not os.path.exists(cwd):
            raise CommandError("Directory not found: %s" % cwd)

        with local.cwd(cwd or local.cwd):
            flat_args = root_node.flatten()
            func = local[flat_args[0]][flat_args[1:]]
            variable_map = cls.get_config('/ENVIRONMENT/variable_map', {})
            with local.env(**variable_map):
                if capture:
                    return func()
                try:
                    func & FG
                    return True
                except ProcessExecutionError:
                    if not quiet:
                        print(
                            "\nDodo Commands error while running this command:"
                        )
                        print("\n\n%s" % func)
                    return False

    # add alias for the legacy 'runcmd' method
    runcmd = run
